from copy import copy
from inspect import getmembers
from warnings import warn

from makefun import wraps, with_signature

try:
    from inspect import signature, Parameter, Signature
except ImportError:
    from funcsigs import signature, Parameter, Signature

try:
    from typing import Any, Tuple, Callable, Union, TypeVar
    try:
        from typing import Type
    except ImportError:
        pass
    T = TypeVar('T')
except ImportError:
    pass

from decopatch import DECORATED, function_decorator, class_decorator

from autoclass.utils import is_attr_selected, validate_include_exclude
from autoclass.utils import get_constructor
from autoclass.utils import _check_known_decorators


__GETTER_OVERRIDE_ANNOTATION = '__getter_override__'
__SETTER_OVERRIDE_ANNOTATION = '__setter_override__'


class IllegalGetterSignatureException(Exception):
    """ This is raised whenever an overridden getter has an illegal signature"""


class IllegalSetterSignatureException(Exception):
    """ This is raised whenever an overridden setter has an illegal signature"""


class DuplicateOverrideError(Exception):
    """ This is raised whenever a getter or setter is overridden twice for the same attribute"""


@class_decorator
def autoprops(include=None,  # type: Union[str, Tuple[str]]
              exclude=None,  # type: Union[str, Tuple[str]]
              cls=DECORATED):
    """
    A decorator to automatically generate all properties getters and setters from the class constructor.
    * if a @contract annotation exist on the __init__ method, mentioning a contract for a given parameter, the
    parameter contract will be added on the generated setter method
    * The user may override the generated getter and/or setter by creating them explicitly in the class and annotating
    them with @getter_override or @setter_override. Note that the contract will still be dynamically added on the
    setter, even if the setter already has one (in such case a `UserWarning` will be issued)

    :param include: a tuple of explicit attribute names to include (None means all)
    :param exclude: a tuple of explicit attribute names to exclude. In such case, include should be None.
    :return:
    """
    return autoprops_decorate(cls, include=include, exclude=exclude)


def autoprops_decorate(cls,           # type: Type[T]
                       include=None,  # type: Union[str, Tuple[str]]
                       exclude=None   # type: Union[str, Tuple[str]]
                       ):
    # type: (...) -> Type[T]
    """
    To automatically generate all properties getters and setters from the class constructor manually, without using
    @autoprops decorator.
    * if a @contract annotation exist on the __init__ method, mentioning a contract for a given parameter, the
    parameter contract will be added on the generated setter method
    * The user may override the generated getter and/or setter by creating them explicitly in the class and annotating
    them with @getter_override or @setter_override. Note that the contract will still be dynamically added on the
    setter, even if the setter already has one (in such case a `UserWarning` will be issued)

    :param cls: the class on which to execute. Note that it won't be wrapped.
    :param include: a tuple of explicit attribute names to include (None means all)
    :param exclude: a tuple of explicit attribute names to exclude. In such case, include should be None.
    :return:
    """

    # first check that we do not conflict with other known decorators
    _check_known_decorators(cls, '@autoprops')

    # perform the class mod
    _execute_autoprops_on_class(cls, include=include, exclude=exclude)

    # TODO better create a wrapper than modify the class? Probably not
    # class Autoprops_Wrapper(object):
    #     def __init__(self, *args, **kwargs):
    #         self.wrapped = cls(*args, **kwargs)
    #
    # return Autoprops_Wrapper

    return cls


def _execute_autoprops_on_class(object_type,   # type: Type[T]
                                include=None,  # type: Union[str, Tuple[str]]
                                exclude=None   # type: Union[str, Tuple[str]]
                                ):
    """
    This method will automatically add one getter and one setter for each constructor argument, except for those
    overridden using autoprops_override_decorate(), @getter_override or @setter_override.
    It will add a @contract on top of all setters (generated or overridden, if they don't already have one)

    :param object_type: the class on which to execute.
    :param include: a tuple of explicit attribute names to include (None means all)
    :param exclude: a tuple of explicit attribute names to exclude. In such case, include should be None.
    :return:
    """
    # 0. first check parameters
    validate_include_exclude(include, exclude)

    # 1. Find the __init__ constructor signature and possible pycontracts @contract
    constructor = get_constructor(object_type, allow_inheritance=True)
    s = signature(constructor)

    # option a) pycontracts
    contracts_dict = constructor.__contracts__ if hasattr(constructor, '__contracts__') else {}
    # option b) valid8
    validators_dict = constructor.__validators__ if hasattr(constructor, '__validators__') else {}

    # 2. For each attribute that is not 'self' and is included and not excluded, add the property
    added = []
    for attr_name in s.parameters.keys():
        if is_attr_selected(attr_name, include=include, exclude=exclude):
            added.append(attr_name)

            # pycontract
            if attr_name in contracts_dict.keys():
                pycontract = contracts_dict[attr_name]
            else:
                pycontract = None

            # valid8 validators: create copies, because we will modify them (changing the validated function ref)
            if attr_name in validators_dict.keys():
                validators = [copy(v) for v in validators_dict[attr_name]]
            else:
                validators = None

            _add_property(object_type, s.parameters[attr_name], pycontract=pycontract, validators=validators)

    # 3. Finally check that there is no overridden setter or getter that does not correspond to an attribute
    extra_overrides = getmembers(object_type, predicate=(lambda fun: callable(fun) and
                                                                     (hasattr(fun, __GETTER_OVERRIDE_ANNOTATION)
                                                                      and getattr(fun, __GETTER_OVERRIDE_ANNOTATION) not in added)
                                                                      or
                                                                     (hasattr(fun, __SETTER_OVERRIDE_ANNOTATION))
                                                                      and getattr(fun, __SETTER_OVERRIDE_ANNOTATION) not in added)
                                 )
    if len(extra_overrides) > 0:
        raise AttributeError('Attribute named \'' + extra_overrides[0][0] + '\' was not found in constructor signature.'
                             'Therefore its getter/setter can not be overridden by function '
                             + extra_overrides[0][1].__qualname__)


def _add_property(object_type,      # type: Type[T]
                  parameter,        # type: Parameter
                  pycontract=None,  # type: Any
                  validators=None   # type: Any
                  ):
    """
    A method to dynamically add a property to a class with the optional given pycontract or validators.
    If the property getter and/or setter have been overridden, it is taken into account too.

    :param object_type: the class on which to execute.
    :param parameter:
    :param pycontract:
    :param validators:
    :return:
    """
    property_name = parameter.name

    # 1. create the private field name , e.g. '_foobar'
    private_property_name = '_' + property_name

    # 2. property getter (@property) - create or use overridden
    getter_fun = _get_getter_fun(object_type, parameter, private_property_name)

    # 3. property setter (@property_name.setter) - create or use overridden
    setter_fun, var_name = _get_setter_fun(object_type, parameter, private_property_name)

    # 4. add the contract to the setter, if any
    setter_fun_with_possible_contract = setter_fun
    if pycontract is not None:
        setter_fun_with_possible_contract = _add_contract_to_setter(setter_fun, var_name, pycontract,
                                                                    property_name)
    elif validators is not None:
        setter_fun_with_possible_contract = _add_validators_to_setter(setter_fun, var_name, validators,
                                                                      property_name)

    # 5. change the function name to make it look nice
    setter_fun_with_possible_contract.__name__ = property_name
    setter_fun_with_possible_contract.__module__ = object_type.__module__
    setter_fun_with_possible_contract.__qualname__ = object_type.__name__ + '.' + property_name
    # __annotations__
    # __doc__
    # __dict__

    # 6. Finally add the property to the class
    # WARNING : property_obj.setter(f) does absolutely nothing :) > we have to assign the result
    # setattr(object_type, property_name, property_obj.setter(f))
    new_prop = property(fget=getter_fun, fset=setter_fun_with_possible_contract)

    # # specific for enforce: here we might wrap the overriden property setter on which enforce has already written
    # # something.
    # if hasattr(setter_fun_with_possible_contract, '__enforcer__'):
    #     new_prop.__enforcer__ = setter_fun_with_possible_contract.__enforcer__
    # DESIGN DECISION > although this would probably work, it is probably better to 'force' users to always use the
    # @autoprops annotation BEFORE any other annotation. This is now done in autoprops_decorate

    setattr(object_type, property_name, new_prop)


def _has_annotation(annotation, value):
    """ Returns a function that can be used as a predicate in get_members, that  """

    def matches_property_name(fun):
        """ return true if fun is a callable that has the correct annotation with value """
        return callable(fun) and hasattr(fun, annotation) \
               and getattr(fun, annotation) is value
    return matches_property_name


def _get_getter_fun(object_type,           # type: Type
                    parameter,             # type: Parameter
                    private_property_name  # type: str
                    ):
    """
    Utility method to find the overridden getter function for a given property, or generate a new one

    :param object_type:
    :param property_name:
    :param private_property_name:
    :return:
    """

    property_name = parameter.name

    # -- check overridden getter for this property name
    overridden_getters = getmembers(object_type, predicate=_has_annotation(__GETTER_OVERRIDE_ANNOTATION, property_name))

    if len(overridden_getters) > 0:
        if len(overridden_getters) > 1:
            raise DuplicateOverrideError('Getter is overridden more than once for attribute name : ' + property_name)

        # --use the overridden getter
        getter_fun = overridden_getters[0][1]

        # --check its signature
        s = signature(getter_fun)
        if not ('self' in s.parameters.keys() and len(s.parameters.keys()) == 1):
            raise IllegalGetterSignatureException('overridden getter must only have a self parameter, found ' +
                                                  str(len(s.parameters.items()) - 1) + ' for function ' + str(
                getter_fun.__qualname__))

        # --use the overridden getter ?
        property_obj = property(getter_fun)
    else:
        # -- generate the getter :
        def autoprops_generated_getter(self):
            return getattr(self, private_property_name)

        # -- use the generated getter
        getter_fun = autoprops_generated_getter

        try:
            annotations = getter_fun.__annotations__
        except AttributeError:
            # python 2
            pass
        else:
            annotations['return'] = parameter.annotation  # add type hint to output declaration

    return getter_fun


def _get_setter_fun(object_type,           # type: Type
                    parameter,             # type: Parameter
                    private_property_name  # type: str
                    ):
    """
    Utility method to find the overridden setter function for a given property, or generate a new one

    :param object_type:
    :param property_name:
    :param property_type:
    :param private_property_name:
    :return:
    """
    # the property will have the same name than the constructor argument
    property_name = parameter.name

    overridden_setters = getmembers(object_type, _has_annotation(__SETTER_OVERRIDE_ANNOTATION, property_name))

    if len(overridden_setters) > 0:
        # --check that we only have one
        if len(overridden_setters) > 1:
            raise DuplicateOverrideError('Setter is overridden more than once for attribute name : %s' % property_name)

        # --use the overridden setter
        setter_fun = overridden_setters[0][1]

        try:
            # python 2
            setter_fun = setter_fun.im_func
        except AttributeError:
            pass

        # --find the parameter name and check the signature
        s = signature(setter_fun)
        p = [attribute_name for attribute_name, param in s.parameters.items() if attribute_name is not 'self']
        if len(p) != 1:
            try:
                qname = setter_fun.__qualname__
            except AttributeError:
                qname = setter_fun.__name__
            raise IllegalSetterSignatureException('overridden setter must have only 1 non-self argument, found ' +
                                                  '%s for function %s'
                                                  '' % (len(s.parameters.items()) - 1, qname))
        var_name = p[0]

    else:
        # --create the setter, equivalent of:
        # ** Dynamically compile a wrapper with correct argument name **
        sig = Signature(parameters=[Parameter('self', kind=Parameter.POSITIONAL_OR_KEYWORD), parameter])

        @with_signature(sig)
        def autoprops_generated_setter(self, **kwargs):
            setattr(self, private_property_name, kwargs.popitem()[1])

        setter_fun = autoprops_generated_setter
        var_name = property_name

    return setter_fun, var_name


def _add_contract_to_setter(setter_fun, var_name, property_contract, property_name):

    # 0. check that we can import contracts
    try:
        # noinspection PyUnresolvedReferences
        from contracts import ContractNotRespected, contract
    except ImportError as e:
        raise Exception('Use of _add_contract_to_setter requires that PyContract library is installed. Check that you '
                        'can \'import contracts\'')

    try:
        # python 2
        setter_fun = setter_fun.im_func
    except AttributeError:
        pass

    # -- check if a contract already exists on the function
    if hasattr(setter_fun, '__contracts__'):
        try:
            qname = str(setter_fun.__qualname__)
        except AttributeError:
            qname = setter_fun.__name__
        msg = "overridden setter for attribute %s implemented by function %s has a contract while there is a " \
              "contract already defined for this property in the __init__ constructor. This will lead to " \
              "double-contract in the final setter, please remove the one on the overridden setter." \
              "" % (property_name, qname)
        warn(msg)

    # -- add the generated contract
    setter_fun_with_possible_contract = contract(setter_fun, **{var_name: property_contract})

    # the only thing we can't do is to replace the function's parameter name dynamically in the error messages
    # so we wrap the function again to catch the potential pycontracts error :(
    @wraps(setter_fun_with_possible_contract)
    def _contracts_parser_interceptor(self, *args, **kwargs):
        try:
            return setter_fun_with_possible_contract(self, *args, **kwargs)
        except ContractNotRespected as er:
            er.error = er.error.replace('\'val\'', '\'' + property_name + '\'')
            raise er

    return _contracts_parser_interceptor


def _add_validators_to_setter(setter_fun, var_name, validators, property_name):

    # 0. check that we can import valid8
    # note: this is useless now but maybe one day validate will be another project ?
    try:
        # noinspection PyUnresolvedReferences
        from valid8 import decorate_with_validators
    except ImportError:
        raise Exception('Use of _add_contract_to_setter requires that valid8 library is installed. Check that you can'
                        ' \'import valid8\'')

    # -- check if a contract already exists on the function
    if hasattr(setter_fun, '__validators__'):
        msg = 'overridden setter for attribute ' + property_name + ' implemented by function ' \
              + str(setter_fun.__qualname__) + ' has validators while there are validators already defined ' \
              'for this property in the __init__ constructor. This will lead to double-contract in the final ' \
              'setter, please remove the one on the overridden setter.'
        warn(msg)

    # -- add the generated contract
    setter_fun_with_validation = decorate_with_validators(setter_fun, **{var_name: validators})

    # bind the validators to the setter function so that error message is correct
    for v in validators:
        v.validated_func = setter_fun_with_validation

    # # the only thing we can't do is to replace the function's parameter name dynamically in the validation error
    # #  messages so we wrap the function again to catch the potential pycontracts error :(
    # # old:
    # # @functools.wraps(func) -> to make the wrapper function look like the wrapped function
    # # def wrapper(self, *args, **kwargs):
    # # new:
    # # we now use 'decorate' to have a wrapper that has the same signature, see below
    # def _contracts_parser_interceptor(func, self, *args, **kwargs):
    #     try:
    #         return func(self, *args, **kwargs)
    #     except ContractNotRespected as e:
    #         e.error = e.error.replace('\'val\'', '\'' + property_name + '\'')
    #         raise e

    # f = _contracts_parser_interceptor(f)
    # setter_fun_with_possible_contract = decorate(setter_fun_with_possible_contract, _contracts_parser_interceptor)
    return setter_fun_with_validation


@function_decorator
def getter_override(attribute=None,  # type: str
                    f=DECORATED
                    ):
    """
    A decorator to indicate an overridden getter for a given attribute. If the attribute name is None, the function name
    will be used as the attribute name.

    :param attribute: the attribute name for which the decorated function is an overridden getter
    :return:
    """
    return autoprops_override_decorate(f, attribute=attribute, is_getter=True)


@function_decorator
def setter_override(attribute=None,  # type: str
                    f=DECORATED
                    ):
    """
    A decorator to indicate an overridden setter for a given attribute. If the attribute name is None, the function name
     will be used as the attribute name. The @contract will still be dynamically added.

    :param attribute: the attribute name for which the decorated function is an overridden setter
    :return:
    """
    return autoprops_override_decorate(f, attribute=attribute, is_getter=False)


def autoprops_override_decorate(func,            # type: Callable
                                attribute=None,  # type: str
                                is_getter=True   # type: bool
                                ):
    # type: (...) -> Callable
    """
    Used to decorate a function as an overridden getter or setter, without using the @getter_override or
    @setter_override annotations. If the overridden setter has no @contract, the contract will still be
     dynamically added. Note: this should be executed BEFORE @autoprops or autoprops_decorate().

    :param func: the function on which to execute. Note that it won't be wrapped but simply annotated.
    :param attribute: the attribute name. If None, the function name will be used
    :param is_getter: True for a getter override, False for a setter override.
    :return:
    """

    # Simply annotate the fact that this is a function
    attr_name = attribute or func.__name__
    if is_getter:
        if hasattr(func, __GETTER_OVERRIDE_ANNOTATION):
            raise DuplicateOverrideError('Getter is overridden twice for attribute name : ' + attr_name)
        else:
            # func.__getter_override__ = attr_name
            setattr(func, __GETTER_OVERRIDE_ANNOTATION, attr_name)
    else:
        if hasattr(func, __SETTER_OVERRIDE_ANNOTATION):
            raise DuplicateOverrideError('Setter is overridden twice for attribute name : ' + attr_name)
        else:
            # func.__setter_override__ = attr_name
            setattr(func, __SETTER_OVERRIDE_ANNOTATION, attr_name)

    return func



from random import random

import pytest
try:
    from typing import Tuple, Dict
except ImportError:
    pass

from autoclass import autohash, autoclass


@pytest.mark.parametrize('only_public_fields', [True, False], ids=lambda x: 'only_public' if x else 'including class-private dunder fields')
@pytest.mark.parametrize('only_constructor_args', [True, False], ids=lambda x: 'only_constructor_args' if x else 'all_obj_fields')
def test_autohash(only_constructor_args, only_public_fields):
    """ @autohash functionality with various customization options for only_constructor_args/only_public_fields """

    @autohash(only_constructor_args=only_constructor_args, only_public_fields=only_public_fields)
    class FooConfigA(object):

        dummy_class_field = 'just to be sure it does not appear'

        def __init__(self,
                     a,  # type: str
                     b   # type: Tuple[str, str]
                     ):
            self.a = a
            self.b = b
            self.c = 1
            self._weak_private = 2
            self.__class_private = 3

        def dummy_func(self):
            """ we create this just to be sure the function is not in the hash """
            pass

    # *** first object 'a'
    a = FooConfigA('rhubarb', ('pie', 'pie2'))
    a.new_field = 1
    a._new_field_weak_private = 2
    a.__new_field_class_private_incorrect = 3

    class Dummy:
        a.__new_field_class_private = 4

    # *** b is fully identical to a (constructor args + static/dynamic public/private fields)
    b = FooConfigA('rhubarb', ('pie', 'pie2'))
    b.new_field = 1
    b._new_field_weak_private = 2
    b.__new_field_class_private_incorrect = 3

    class Dummy:
        b.__new_field_class_private = 4

    # *** d is identical to a but only for constructor args + public fields
    d = FooConfigA('rhubarb', ('pie', 'pie2'))
    d._weak_private = random()
    d.__class_private = random()
    d.new_field = 1
    d._new_field_weak_private = random()
    d.__new_field_class_private_incorrect = random()

    class Dummy2:
        d.__new_field_class_private = random()

    # *** d is identical to a but only for constructor args
    e = FooConfigA('rhubarb', ('pie', 'pie2'))

    # *** e is different from a
    f = FooConfigA('rhubarb', ('pie3', ''))

    # check that the hash works
    assert hash(a) == hash(b)

    if only_constructor_args:
        assert hash(a) == hash(d)
        assert hash(a) == hash(e)
    elif only_public_fields:
        assert hash(a) == hash(d)
        assert hash(a) != hash(e)
    else:
        assert hash(a) != hash(d)
        assert hash(a) != hash(e)

    assert hash(a) != hash(f)


def test_autohash_exclude():
    """ Tests that exclusion works correctly with autohash """

    @autoclass(autohash=False)
    @autohash(exclude='bar')  # we have to put an underscore because that's the property
    class Foo(object):
        def __init__(self,
                     foo,  # type: str
                     bar   # type: Dict[str, str]
                     ):
            pass

    a = Foo('hello', dict())
    assert hash(a) == hash((a.foo, ))  # supposed to work since we exclude the dict (unhashable)

    # combined tests (foo is transformed to a property)
    @autoclass(include='foo')
    class Foo(object):
        def __init__(self,
                     foo,  # type: str
                     bar  # type: Dict[str, str]
                     ):
            pass

    a = Foo('hello', dict())
    assert hash(a) == hash((a.foo, ))  # supposed to work since we exclude the dict (unhashable)

    @autoclass(exclude='bar')
    class Foo(object):
        def __init__(self,
                     foo,  # type: str
                     bar  # type: Dict[str, str]
                     ):
            pass

    a = Foo('hello', dict())
    assert hash(a) == hash((a.foo,))  # supposed to work since we exclude the dict (unhashable)



import sys

import pytest

try:
    from typing import List
except ImportError:
    pass
from valid8 import Boolean, minlens, validate_io, ValidationError, gt, minlen


from autoclass import autoargs, autoprops, autodict, autoclass, autohash


def test_readme_usage_autoprops_validate():
    @autoprops
    class FooConfigA(object):
        @autoargs
        @validate_io(a=minlens(0))
        def __init__(self,
                     a  # type: str
                     ):
            pass

    t = FooConfigA('rhubarb')

    # check that the generated getters work
    t.a = 'r'
    assert t.a == 'r'

    # check that there are validators on the generated setters
    with pytest.raises(ValidationError):
        t.a = ''  # raises ValidationError


def test_readme_usage_autodict_1():
    """ basic autodict without and with autoargs """

    # ** without autoargs
    @autodict
    class A(object):
        def __init__(self,
                     a,  # type: int
                     b   # type: str
                     ):
            self.a = a
            self.b = b

    o = A(1, 'r')
    # o behaves like a read-only dict
    assert o == dict(o)
    assert o == {'a': 1, 'b': 'r'}

    # you can create an object from a dict too thanks to the generated class function
    p = A.from_dict({'a': 1, 'b': 'r'})
    assert p == o

    # str and repr methods show interesting stuff
    str(p)  # "A({'a': 1, 'b': 'r'})"
    repr(p)  # "A({'a': 1, 'b': 'r'})"

    # ** with autoargs
    @autodict
    class B(object):
        @autoargs
        def __init__(self,
                     a,  # type: int
                     b  # type: str
                     ):
            pass

    o = B(1, 'r')
    # same results
    assert o == {'a': 1, 'b': 'r'}

    # you can create an object from a dict too thanks to the generated class function
    p = B.from_dict({'a': 1, 'b': 'r'})
    assert p == o


def test_readme_usage_autodict_2():
    """ basic autodict with other and private fields """

    @autodict
    class C(object):
        @autoargs
        def __init__(self,
                     a,  # type: str
                     b   # type: List[str]
                     ):
            self.non_constructor_arg = 't'
            self._private = 1
            self.__class_private = 't'

    o = C(1, 'r')
    # only fields corresponding to constructor arguments are visible
    assert o == {'a': 1, 'b': 'r'}


def test_readme_usage_autodict_3():

    @autodict(only_constructor_args=False, only_public_fields=False)
    class D(object):
        @autoargs
        def __init__(self,
                     a,  # type: str
                     b  # type: List[str]
                     ):
            self.non_constructor_arg = 'b'
            self._private = 1
            self.__class_private = 't'

    o = D(1, 'r')
    # o behaves like a read-only dict, all fields are now visible
    assert o == dict(o)
    assert o == {'a': 1, 'b': 'r',
                 'non_constructor_arg': 'b',
                 '_private': 1,
                 '_D__class_private': 't'}  # notice the name


def test_readme_usage_autohash_1():
    @autohash
    class A(object):
        def __init__(self,
                     a,  # type: int
                     b   # type: str
                     ):
            self.a = a
            self.b = b

    o = A(1, 'r')
    o._test = 2

    # o is hashable
    # the order of vars(self).values() does not seem stable across CPython versions: skipping this test
    # assert hash(o) == hash((1, 'r', 2))

    p = A(1, 'r')
    p._test = 2
    # o and p have identical hash
    assert hash(o) == hash(p)

    # dynamic and private fields are taken into account by default
    p._test = 3
    assert hash(o) != hash(p)


def test_readme_usage_autohash_2():

    from random import random

    @autohash(only_constructor_args=True, only_public_fields=True)
    class D(object):
        @autoargs
        def __init__(self,
                     a,  # type: str
                     _b  # type: str
                     ):
            self.non_constructor_arg = random()
            self._private = random()
            self.__class_private = random()

    o = D(1, 'r')
    p = D(1, 'r')

    # o and p have the same hash because only the constructor arguments are taken into account
    assert hash(o) == hash(p)
    assert hash(o) == hash((1, 'r'))


@pytest.mark.skipif(sys.version_info < (3, 0), reason="type hints do not work in python 2")
@pytest.mark.skipif(sys.version_info >= (3, 7), reason="enforce does not work correctly under python 3.7+")
def test_readme_usage_autoclass():

    from ._tests_pep484 import test_readme_usage_autoclass
    test_readme_usage_autoclass()


@pytest.mark.skipif(sys.version_info < (3, 0), reason="type hints do not work in python 2")
@pytest.mark.skipif(sys.version_info >= (3, 7), reason="enforce does not work correctly under python 3.7+")
def test_readme_usage_autoclass_custom():

    from ._tests_pep484 import test_readme_usage_autoclass_custom
    test_readme_usage_autoclass_custom()



from collections import Sequence
from valid8 import validate

try:  # python 3.5+
    from typing import Union, Tuple
except ImportError:
    pass


def validate_include_exclude(include, exclude):
    """
    Common validator for include and exclude arguments
    :param include:
    :param exclude:
    :return:
    """
    if include is not None and exclude is not None:
        raise ValueError("Only one of 'include' or 'exclude' argument should be provided.")
    validate('include', include, instance_of=(str, Sequence), enforce_not_none=False)
    validate('exclude', exclude, instance_of=(str, Sequence), enforce_not_none=False)


def is_attr_selected(attr_name,     # type: str
                     include=None,  # type: Union[str, Tuple[str]]
                     exclude=None   # type: Union[str, Tuple[str]]
                     ):
    """decide whether an action has to be performed on the attribute or not, based on its name"""

    if include is not None and exclude is not None:
        raise ValueError('Only one of \'include\' or \'exclude\' argument should be provided.')

    # win time by not doing this
    # check_var(include, var_name='include', var_types=[str, tuple], enforce_not_none=False)
    # check_var(exclude, var_name='exclude', var_types=[str, tuple], enforce_not_none=False)

    if attr_name is 'self':
        return False
    if exclude and attr_name in exclude:
        return False
    if not include or attr_name in include:
        return True
    else:
        return False


def get_constructor(typ,
                    allow_inheritance=False  # type: bool
                    ):
    """
    Utility method to return the unique constructor (__init__) of a type

    :param typ: a type
    :param allow_inheritance: if True, the constructor will be returned even if it is not defined in this class
    (inherited). By default this is set to False: an exception is raised when no constructor is explicitly defined in
    the class
    :return: the found constructor
    """
    if allow_inheritance:
        return typ.__init__
    else:
        # check that the constructor is really defined here
        if '__init__' in typ.__dict__:
            return typ.__init__
        else:
            raise Exception('No explicit constructor was found for class ' + str(typ))


class AutoclassDecorationException(Exception):
    pass


def _check_known_decorators(typ, calling_decorator  # type: str
                            ):
    # type: (...) -> bool
    """
    Checks that a given type is not already decorated by known decorators that may cause trouble.
    If so, it raises an Exception
    :return:
    """
    for member in typ.__dict__.values():
        if hasattr(member, '__enforcer__'):
            raise AutoclassDecorationException('It seems that @runtime_validation decorator was applied to type <'
                                               + str(typ) + '> BEFORE ' + calling_decorator + '. This is not supported '
                                               'as it may lead to counter-intuitive behaviour, please change the order '
                                               'of the decorators on <' + str(typ) + '>')


def method_already_there(object_type, method_name, this_class_only=False):
    """
    Returns True if method `method_name` is already implemented by object_type, that is, its implementation differs from
    the one in `object`.

    :param object_type:
    :param method_name:
    :param this_class_only:
    :return:
    """
    if this_class_only:
        return method_name in vars(object_type)  # or object_type.__dict__
    else:
        try:
            method = getattr(object_type, method_name)
        except AttributeError:
            return False
        else:
            return method is not None and method is not getattr(object, method_name, None)


def possibly_replace_with_property_name(object_type, att_name):
    """
    Returns the attribute name or the corresponding property name if there is one

    :param object_type:
    :param att_name:
    :return:
    """
    return att_name[1:] if is_property_related_attr(object_type, att_name) else att_name


def is_property_related_attr(object_type, att_name):
    """
    Returns True if the attribute name without a leading underscore corresponds to a property name in that class
    :param object_type:
    :param att_name:
    :return:
    """
    return att_name[0] == '_' and isinstance(getattr(object_type, att_name[1:], None), property)



import sys
import pytest

from autoclass import autoclass


@pytest.mark.skipif(sys.version_info < (3, 0), reason="type hints do not work in python 2")
@pytest.mark.skipif(sys.version_info >= (3, 7), reason="enforce does not work correctly under python 3.7+")
def test_autoclass_enforce_validate_not_reversed():
    """ Tests that if we reverse the annotations orders, it still works. Currently it fails """

    from ._tests_pep484 import test_autoclass_enforce_validate_not_reversed
    test_autoclass_enforce_validate_not_reversed()


@pytest.mark.skipif(sys.version_info < (3, 0), reason="type hints do not work in python 2")
@pytest.mark.skipif(sys.version_info >= (3, 7), reason="enforce does not work correctly under python 3.7+")
def test_autoclass_enforce_validate_reversed():
    """ Tests that if we reverse the annotations orders, it still works. Currently it fails """

    from ._tests_pep484 import test_autoclass_enforce_validate_reversed
    test_autoclass_enforce_validate_reversed()


@pytest.mark.skipif(sys.version_info < (3, 0), reason="type hints do not work in python 2")
def test_readme_pytypes():
    """ Makes sure that the code in the documentation page is correct for the pytypes example """

    from ._tests_pep484 import test_readme_pytypes
    test_readme_pytypes()


@pytest.mark.skipif(sys.version_info < (3, 0), reason="type hints do not work in python 2")
@pytest.mark.skipif(sys.version_info >= (3, 7), reason="enforce does not work correctly under python 3.7+")
def test_readme_enforce():
    """ Makes sure that the code in the documentation page is correct for the enforce example """

    from ._tests_pep484 import test_readme_enforce
    test_readme_enforce()


def test_autoclass_inheritance():

    @autoclass
    class Foo(object):
        def __init__(self, foo1, foo2=0):
            pass

    @autoclass
    class Bar(Foo):
        def __init__(self, bar, foo1, foo2=0):
            # this constructor is not actually needed in this case since all fields are already self-assigned here
            super(Bar, self).__init__(foo1, foo2)
            # pass

    a = Bar(2, 'th')
    assert a == {'bar': 2, 'foo1': 'th', 'foo2': 0}
    assert a['foo1'] == 'th'

    # iteration order is fixed
    assert list(a.keys()) == ['bar', 'foo1', 'foo2']

    # order in prints is fixed
    assert str(a) == "Bar({'bar': 2, 'foo1': 'th', 'foo2': 0})"



import sys

import pytest

from numbers import Real, Integral
from valid8 import Boolean, validate, validate_arg, instance_of, is_multiple_of, InputValidationError

try:
    from typing import Optional
except ImportError:
    pass

from autoclass import autoargs, autoprops, setter_override, autoclass


def test_readme_index_basic():
    """ First basic example in the doc """

    @autoclass
    class House(object):
        def __init__(self, name, nb_floors=1):
            pass

    a = House('my_house', 3)
    assert str(a) == "House({'name': 'my_house', 'nb_floors': 3})"
    assert [att for att in a.keys()] == ['name', 'nb_floors']
    assert {a, a} == {a}
    assert a == {'name': 'my_house', 'nb_floors': 3}


def test_readme_index_basic2():
    """ Second basic example in the doc: adding setter override """

    global t
    t = ''

    @autoclass
    class House(object):
        def __init__(self, name, nb_floors=1):
            pass

        @setter_override
        def nb_floors(self, nb_floors=1):
            global t
            t = 'Set nb_floors to {}'.format(nb_floors)
            self._nb_floors = nb_floors

    assert t == ''

    obj = House('my_house')
    assert t == 'Set nb_floors to 1'

    obj.nb_floors = 3
    assert t == 'Set nb_floors to 3'


@pytest.mark.skipif(sys.version_info < (3, 0), reason="type hints do not work in python 2")
@pytest.mark.skipif(sys.version_info >= (3, 7), reason="enforce does not work correctly under python 3.7+")
def test_readme_enforce_simple():
    """ Example in the doc with enforce """

    from ._tests_pep484 import test_readme_enforce_simple
    test_readme_enforce_simple()


@pytest.mark.skipif(sys.version_info < (3, 0), reason="type hints do not work in python 2")
def test_readme_index_pytypes_simple():
    """ Example in the doc with pytypes """

    from ._tests_pep484 import test_readme_index_pytypes_simple
    test_readme_index_pytypes_simple()


def test_readme_index_valid8_simple():
    """ Example in the doc with valid8 """

    from mini_lambda import s, x, Len

    # Here we define our 2 validation errors
    class InvalidNameError(InputValidationError):
        help_msg = 'name should be a non-empty string'

    class InvalidSurfaceError(InputValidationError):
        help_msg = 'Surface should be between 0 and 10000 and be a multiple of 100.'

    @autoclass
    class House(object):

        @validate_arg('name', instance_of(str), Len(s) > 0,
                      error_type=InvalidNameError)
        @validate_arg('surface', (x >= 0) & (x < 10000), is_multiple_of(100),
                      error_type=InvalidSurfaceError)
        def __init__(self, name, surface=None):
            pass

    obj = House('sweet home', 200)

    obj.surface = None  # Valid (surface is nonable by signature)
    with pytest.raises(InvalidNameError):
        obj.name = 12  # InvalidNameError
    with pytest.raises(InvalidSurfaceError):
        obj.surface = 10000  # InvalidSurfaceError


@pytest.mark.skipif(sys.version_info < (3, 0), reason="type hints do not work in python 2")
@pytest.mark.skipif(sys.version_info >= (3, 7), reason="enforce does not work correctly under python 3.7+")
def test_readme_index_enforce_valid8():
    """ Makes sure that the code in the documentation page is correct for the enforce + valid8 example """

    from ._tests_pep484 import test_readme_index_enforce_valid8
    test_readme_index_enforce_valid8()


def test_readme_pycontracts_simple():
    """ Simple test with pycontracts """

    from contracts import contract, ContractNotRespected

    @autoclass
    class House(object):

        @contract(name='str[>0]',
                  surface='None|(int,>=0,<10000)')
        def __init__(self, name, surface):
            pass

    obj = House('sweet home', 200)

    obj.surface = None  # Valid (surface is nonable by signature)
    with pytest.raises(ContractNotRespected):
        obj.name = ''  # InvalidNameError
    with pytest.raises(ContractNotRespected):
        obj.surface = 10000  # InvalidSurfaceError


def test_readme_old_way():
    """ Makes sure that the code in the documentation page is correct for the 'old way' of writing classes """

    class HouseConfiguration(object):
        def __init__(self,
                     name,               # type: str
                     surface,            # type: Real
                     nb_floors=1,        # type: Optional[Integral]
                     with_windows=False  # type: Boolean
                     ):
            self.name = name
            self.surface = surface
            self.nb_floors = nb_floors
            self.with_windows = with_windows

        # --name
        @property
        def name(self):
            return self._name

        @name.setter
        def name(self,
                 name  # type: str
                 ):
            validate('name', name, instance_of=str)
            self._name = name

        # --surface
        @property
        def surface(self):
            # type: (...) -> Real
            return self._surface

        @surface.setter
        def surface(self,
                    surface  # type: Real
                    ):
            validate('surface', surface, instance_of=Real, min_value=0, min_strict=True)
            self._surface = surface

        # --nb_floors
        @property
        def nb_floors(self):
            # type: (...) -> Optional[Integral]
            return self._nb_floors

        @nb_floors.setter
        def nb_floors(self,
                      nb_floors  # type: Optional[Integral]
                      ):
            validate('nb_floors', nb_floors, instance_of=Integral, enforce_not_none=False)
            self._surface = nb_floors  # !**

        # --with_windows
        @property
        def with_windows(self):
            # type: (...) -> Boolean
            return self._with_windows

        @with_windows.setter
        def with_windows(self,
                         with_windows  # type: Boolean
                         ):
            validate('with_windows', with_windows, instance_of=Boolean)
            self._with_windows = with_windows

    HouseConfiguration('test', 0.1)


def test_readme_pycontracts_complex():
    """ Makes sure that the code in the documentation page is correct for the PyContracts example """

    from contracts import contract, ContractNotRespected

    @autoprops
    class HouseConfiguration(object):
        @autoargs
        @contract(name='str[>0]',
                  surface='(int|float),>=0',
                  nb_floors='None|int',
                  with_windows='bool')
        def __init__(self,
                     name,  # type: str
                     surface,  # type: Real
                     nb_floors=1,  # type: Optional[Integral]
                     with_windows=False  # type: Boolean
                     ):
            pass

        # -- overriden setter for surface - no need to repeat the @contract
        @setter_override
        def surface(self,
                    surface  # type: Real
                    ):
            assert surface > 0
            self._surface = surface

    t = HouseConfiguration('test', 0.1)
    t.nb_floors = None
    with pytest.raises(ContractNotRespected):
        t.nb_floors = 2.2
    with pytest.raises(ContractNotRespected):
        t.surface = -1


@pytest.mark.skipif(sys.version_info < (3, 0), reason="type hints do not work in python 2")
@pytest.mark.skip(reason="open bug in pytypes https://github.com/Stewori/pytypes/issues/19")
def test_readme_pytypes_validate_complex():
    """ A more complex pytypes + valid8 example """

    from ._tests_pep484 import test_readme_pytypes_validate_complex
    test_readme_pytypes_validate_complex()



from numbers import Real, Integral
from typing import Optional, List

import pytest

from autoclass import setter_override, autoclass, AutoclassDecorationException, autoprops, autoargs
from valid8 import Boolean, ValidationError, validate_io, minlens, gts, between, minlen, gt


def test_autoclass_enforce_validate_not_reversed():
    """"""

    from enforce import runtime_validation, config
    config(dict(mode='covariant'))  # to accept subclasses in validation

    @runtime_validation
    @autoclass
    class HouseConfiguration(object):
        def __init__(self, surface: Real):
            pass

        # -- overriden setter for surface
        @setter_override
        def surface(self, surface):
            print('Set surface to {}'.format(surface))
            self._surface = surface

    t = HouseConfiguration(12)


def test_autoclass_enforce_validate_reversed():
    """"""

    from enforce import runtime_validation, config
    config(dict(mode='covariant'))  # to accept subclasses in validation

    with pytest.raises(AutoclassDecorationException):
        @autoclass
        @runtime_validation
        class HouseConfiguration(object):
            def __init__(self, surface: Real):
                pass

            # -- overriden setter for surface
            @setter_override
            def surface(self, surface):
                print('Set surface to {}'.format(surface))
                self._surface = surface

        t = HouseConfiguration(12)


def test_readme_pytypes():
    from pytypes import typechecked

    @typechecked
    @autoclass
    class HouseConfiguration(object):
        def __init__(self,
                     name: str,
                     surface: Real,
                     nb_floors: Optional[Integral] = 1,
                     with_windows: Boolean = False):
            pass

        # -- overriden setter for surface for custom validation
        @setter_override
        def surface(self, surface):
            assert surface > 0
            self._surface = surface

    t = HouseConfiguration('test', 12, 2)

    # 'Optional' works
    t.nb_floors = None

    # Type validation works
    from pytypes import InputTypeError
    with pytest.raises(InputTypeError):
        t.nb_floors = 2.2

    # Custom validation works
    with pytest.raises(AssertionError):
        t.surface = 0


def test_readme_enforce():
    import enforce as en
    from enforce import runtime_validation

    en.config(dict(mode='covariant'))  # allow subclasses when validating types

    @runtime_validation
    @autoprops
    class HouseConfiguration(object):
        @autoargs
        def __init__(self,
                     name: str,
                     surface: Real,
                     nb_floors: Optional[Integral] = 1,
                     with_windows: Boolean = False):
            pass

        # -- overriden setter for surface for custom validation
        @setter_override
        def surface(self, surface):
            assert surface > 0
            self._surface = surface

    t = HouseConfiguration('test', 12, 2)

    # 'Optional' works
    t.nb_floors = None

    # Type validation works
    from enforce.exceptions import RuntimeTypeError
    with pytest.raises(RuntimeTypeError):
        t.nb_floors = 2.2

    # Custom validation works
    with pytest.raises(AssertionError):
        t.surface = 0


def test_autoprops_enforce_validate():

    import enforce as en
    from enforce import runtime_validation

    en.config(dict(mode='covariant'))  # allow subclasses when validating types

    @runtime_validation
    @autoprops
    class HouseConfiguration(object):
        @autoargs
        @validate_io(name=minlens(0),
                     surface=gts(0),
                     nb_floors=between(1, 100, open_right=True))
        def __init__(self,
                     name: str,
                     surface: Real,
                     nb_floors: Optional[Integral] = 1,
                     with_windows: Boolean = False):
            pass

        # -- overriden setter for surface for custom validation or other things
        @setter_override
        def surface(self, surface):
            print('Set surface to {}'.format(surface))
            self.toto = 'done'
            self._surface = surface

    t = HouseConfiguration('test', 12, 2)

    # 'Optional' works
    t.nb_floors = None

    # Custom print works
    t.surface = 0.1
    assert t.toto == 'done'

    # Type validation works
    from enforce.exceptions import RuntimeTypeError
    with pytest.raises(RuntimeTypeError):
        t.nb_floors = 2.2

    # Custom validation works
    with pytest.raises(ValidationError):
        t.surface = 0


def test_autoprops_enforce_default():
    from enforce import runtime_validation

    @runtime_validation
    @autoprops
    class Foo:
        def __init__(self,
                     mandatory_field: str,
                     optional_field: str = None):
            pass

    a = Foo('t')
    a.optional_field = None  # raises RuntimeTypeError: Argument 'val' was not of type <class 'str'>. Actual type was NoneType.


def test_readme_enforce_simple():
    # we use enforce runtime checker for this example
    from enforce import runtime_validation, config
    config(dict(mode='covariant'))  # to accept subclasses in validation

    @runtime_validation
    @autoclass
    class House:
        def __init__(self, name: str, nb_floors: int = 1):
            pass

    obj = House('my_house')

    from enforce.exceptions import RuntimeTypeError
    with pytest.raises(RuntimeTypeError) as exc_info:
        obj.nb_floors = 'red'
    assert exc_info.value.args[0] == "\n  The following runtime type errors were encountered:\n" \
                                     "       Argument 'nb_floors' was not of type <class 'int'>. " \
                                     "Actual type was str."


def test_readme_index_pytypes_simple():
    from pytypes import typechecked

    @typechecked
    @autoclass
    class House:
        # the constructor below is decorated with PEP484 type hints
        def __init__(self, name: str, nb_floors: int = 1):
            pass

    obj = House('my_house')

    from pytypes import InputTypeError
    with pytest.raises(InputTypeError) as exc_info:
        obj.nb_floors = 'red'
    assert exc_info.value.args[0] == "\n  autoclass.tests._tests_pep484." \
                                     "test_readme_index_pytypes_simple.<locals>.House.nb_floors/setter" \
                                     "\n  called with incompatible types:\nExpected: Tuple[int]\nReceived: Tuple[str]"


def test_readme_index_enforce_valid8():
    # Imports - for type validation
    from numbers import Integral
    from enforce import runtime_validation, config
    config(dict(mode='covariant'))  # type validation will accept subclasses too

    # Imports - for value validation
    from mini_lambda import s, x, Len
    from valid8 import validate_arg, is_multiple_of, InputValidationError

    # 2 custom validation errors for valid8
    class InvalidName(InputValidationError):
        help_msg = 'name should be a non-empty string'

    class InvalidSurface(InputValidationError):
        help_msg = 'Surface should be between 0 and 10000 and be a multiple of 100.'

    @runtime_validation
    @autoclass
    class House:
        @validate_arg('name', Len(s) > 0,
                      error_type=InvalidName)
        @validate_arg('surface', (x >= 0) & (x < 10000), is_multiple_of(100),
                      error_type=InvalidSurface)
        def __init__(self, name: str, surface: Integral = None):
            pass

    obj = House('sweet home', 200)

    obj.surface = None  # Valid (surface is nonable by signature)
    with pytest.raises(InvalidName):
        obj.name = ''  # InvalidNameError
    with pytest.raises(InvalidSurface):
        obj.surface = 10000  # InvalidSurfaceError


def test_readme_pytypes_validate_complex():
    # we use pytypes for this example
    from pytypes import typechecked

    @typechecked
    @autoclass
    class HouseConfiguration(object):

        @validate_io(name=minlens(0),
                     surface=gt(0))
        def __init__(self,
                     name: str,
                     surface: Real,
                     nb_floors: Optional[Integral] = 1,
                     with_windows: Boolean = False):
            pass

        # -- overriden setter for surface for custom validation or other things
        @setter_override
        def surface(self, surface):
            print('Set surface to {}'.format(surface))
            self._surface = surface

    t = HouseConfiguration('test', 12, 2)

    # 'Optional' works
    t.nb_floors = None

    # Type validation works
    from pytypes import InputTypeError
    with pytest.raises(InputTypeError):
        t.nb_floors = 2.2

    # Value validation works
    with pytest.raises(ValidationError):
        t.surface = -1

    # Value validation works in constructor
    with pytest.raises(ValidationError):
        HouseConfiguration('', 12, 2)

    # str and repr work
    assert str(t) == repr(t)
    # "HouseConfiguration({'nb_floors': None, 'with_windows': False, 'surface': 12, 'name': 'test'})"
    # assert eval(repr(t)) == t does not work !

    # dict work
    assert t == {'name': 'test', 'nb_floors': None, 'surface': 12, 'with_windows': False}
    assert t == dict(name='test', nb_floors=None, surface=12, with_windows=False)
    t.keys()
    for k, v in t.items():
        print(str(k) + ': ' + str(v))

    # TODO this is an open bug in pytypes https://github.com/Stewori/pytypes/issues/19
    HouseConfiguration.from_dict({'name': 'test2', 'surface': 1})


def test_readme_usage_autoclass():
    # we will use enforce as the runtime checker
    import enforce as en
    from enforce import runtime_validation
    en.config(dict(mode='covariant'))  # allow subclasses when validating types

    # class definition
    @runtime_validation
    @autoclass
    class AllOfTheAbove:
        @validate_io(a=gt(1), c=minlen(1))
        def __init__(self, a: Integral, b: Boolean, c: Optional[List[str]] = None):
            pass

    # instance creation
    o = AllOfTheAbove(a=2, b=True)

    # @autoargs works
    assert o.a == 2

    # @autoprops works, in combination with any runtime checker (here demonstrated with enforce)
    from enforce.exceptions import RuntimeTypeError
    with pytest.raises(RuntimeTypeError):
        o.b = 1  # RuntimeTypeError Argument 'b' was not of type Boolean. Actual type was int.

    # @autodict works
    assert o == {'a': 2, 'b': True, 'c': None}
    assert AllOfTheAbove.from_dict(o) == o
    assert dict(**o) == o


def test_readme_usage_autoclass_custom():
    # we will use enforce as the runtime checker
    import enforce as en
    from enforce import runtime_validation
    en.config(dict(mode='covariant'))  # allow subclasses when validating types

    # class definition
    @runtime_validation
    @autoclass(autodict=False)
    class PartsOfTheAbove:
        @validate_io(a=gt(1), c=minlen(1))
        def __init__(self, a: Integral, b: Boolean, c: Optional[List[str]] = None):
            pass

    # instance creation
    o = PartsOfTheAbove(a=2, b=True)

    # @autoargs works
    assert o.a == 2

    # @autoprops works, in combination with any runtime checker (here demonstrated with enforce)
    from enforce.exceptions import RuntimeTypeError
    with pytest.raises(RuntimeTypeError):
        o.b = 1  # RuntimeTypeError Argument 'b' was not of type Boolean. Actual type was int.

    # @autodict is disabled
    with pytest.raises(AssertionError):
        assert o == {'a': 2, 'b': True, 'c': None}  # AssertionError
    with pytest.raises(AttributeError):
        assert PartsOfTheAbove.from_dict(
            o) == o  # AttributeError: type object 'PartsOfTheAbove' has no attribute 'from_dict'
    with pytest.raises(TypeError):
        assert dict(**o) == o  # TypeError: type object argument after ** must be a mapping, not PartsOfTheAbove



import sys
from collections import OrderedDict

from makefun import wraps

try:  # python 3+
    from inspect import signature
except ImportError:
    from funcsigs import signature

try:  # python 3.5+
    from typing import Tuple, Callable, Union
except ImportError:
    pass

from decopatch import function_decorator, DECORATED

from autoclass.utils import is_attr_selected, validate_include_exclude


@function_decorator
def autoargs(include=None,  # type: Union[str, Tuple[str]]
             exclude=None,  # type: Union[str, Tuple[str]]
             f=DECORATED
             ):
    """
    Defines a decorator with parameters, to automatically assign the inputs of a function to self PRIOR to executing 
    the function. In other words:
    
    ```
    @autoargs
    def myfunc(a):
        print('hello')
    ```
    
    will create the equivalent of
    
    ```
    def myfunc(a):
        self.a = a
        print('hello')
    ```
    
    Initial code from http://stackoverflow.com/questions/3652851/what-is-the-best-way-to-do-automatic-attribute-assignment-in-python-and-is-it-a#answer-3653049

    :param include: a tuple of attribute names to include in the auto-assignment. If None, all arguments will be
        included by default
    :param exclude: a tuple of attribute names to exclude from the auto-assignment. In such case, include should be None
    :return:
    """
    return autoargs_decorate(f, include=include, exclude=exclude)


def autoargs_decorate(func,          # type: Callable
                      include=None,  # type: Union[str, Tuple[str]]
                      exclude=None   # type: Union[str, Tuple[str]]
                      ):
    # type: (...) -> Callable
    """
    Defines a decorator with parameters, to automatically assign the inputs of a function to self PRIOR to executing 
    the function. This is the inline way to apply the decorator 
    
    ```
    myfunc2 = autoargs_decorate(myfunc)
    ```
    
    
    See autoargs for details.

    :param func: the function to wrap
    :param include: a tuple of attribute names to include in the auto-assignment. If None, all arguments will be
        included by default
    :param exclude: a tuple of attribute names to exclude from the auto-assignment. In such case, include should be None
    :return:
    """

    # (0) first check parameters
    validate_include_exclude(include, exclude)

    # (1) then retrieve function signature
    # attrs, varargs, varkw, defaults = getargspec(func)
    func_sig = signature(func)

    # check that include/exclude dont contain names that are incorrect
    if include is not None:
        incorrect = set([include] if isinstance(include, str) else include) - set(func_sig.parameters.keys())
        if len(incorrect) > 0:
            raise ValueError("@autoargs definition exception: include contains '%s' that is/are "
                             "not part of signature for %s" % (incorrect, func))
    if exclude is not None:
        incorrect = set([exclude] if isinstance(exclude, str) else exclude) - set(func_sig.parameters.keys())
        if len(incorrect) > 0:
            raise ValueError("@autoargs definition exception: exclude contains '%s' that is/are "
                             "not part of signature for %s" % (incorrect, func))

    # TODO this should be in @autoslots decorator at class level, not here.
    # (2) Optionally lock the class only for the provided fields
    # Does not work for the moment. Besides locking fields seems to have issues with pickle serialization
    # so we'd rather not propose this option.
    #
    # See 'attrs' project for this kind of advanced features https://github.com/python-attrs/attrs
    #
    # if lock_class_fields:
    #     if signature_varkw:
    #         raise Exception('cant lock field names with variable kwargs')
    #     else:
    #         object_type = get_class_that_defined_method(func)
    #         if include:
    #             fields = include
    #         else:
    #             fields = signature_attrs[1:]
    #             if signature_varargs:
    #                 fields.append(signature_varargs)
    #             if exclude:
    #                 for a in exclude:
    #                     fields.remove(a)
    #
    #         # right now, doesnot work
    #         _lock_fieldnames_class(object_type, field_names=tuple(fields))

    # (3) Finally, create a wrapper around the function so that all attributes included/not excluded are
    # set to self BEFORE executing the function.
    @wraps(func)
    def init_wrapper(self, *args, **kwargs):

        # match the received arguments with the signature to know who is who, and add default values to get a full list
        bound_values = func_sig.bind(self, *args, **kwargs)
        apply_defaults(bound_values)

        # Assign to self the ones that needs to
        for att_name, att_value in bound_values.arguments.items():
            if is_attr_selected(att_name, include=include, exclude=exclude):
                # value = a normal value, or cur_kwargs as a whole
                setattr(self, att_name, att_value)

        # finally execute the constructor function
        return func(self, *args, **kwargs)

    # return wrapper
    return init_wrapper


if sys.version_info >= (3, 0):
    def apply_defaults(bound_values):
        bound_values.apply_defaults()
else:
    # TODO when funcsigs implements PR https://github.com/aliles/funcsigs/pull/30 remove this
    def apply_defaults(bound_values):
        arguments = bound_values.arguments

        # Creating a new one and not modifying in-place for thread safety.
        new_arguments = []

        for name, param in bound_values._signature.parameters.items():
            try:
                new_arguments.append((name, arguments[name]))
            except KeyError:
                if param.default is not param._empty:
                    val = param.default
                elif param.kind is param._VAR_POSITIONAL:
                    val = ()
                elif param.kind is param._VAR_KEYWORD:
                    val = {}
                else:
                    # BoundArguments was likely created by bind_partial
                    continue
                new_arguments.append((name, val))

        bound_values.arguments = OrderedDict(new_arguments)



from collections import Mapping
from warnings import warn

from six import with_metaclass

try:  # python 3+
    from inspect import signature
except ImportError:
    from funcsigs import signature

try:
    from typing import Any, Tuple, Union, Dict, TypeVar, Callable
    try:
        from typing import Type
    except ImportError:
        pass
    T = TypeVar('T')
except ImportError:
    pass

from autoclass.autoprops_ import DuplicateOverrideError
from autoclass.utils import is_attr_selected, method_already_there, possibly_replace_with_property_name, \
    validate_include_exclude
from autoclass.utils import get_constructor
from autoclass.utils import _check_known_decorators

from decopatch import class_decorator, DECORATED


__AUTODICT_OVERRIDE_ANNOTATION = '__autodict_override__'


@class_decorator
def autodict(include=None,                # type: Union[str, Tuple[str]]
             exclude=None,                # type: Union[str, Tuple[str]]
             only_constructor_args=True,  # type: bool
             only_public_fields=True,     # type: bool
             cls=DECORATED
             ):
    """
    A decorator to makes objects of the class behave like a read-only `dict`. It does several things:
    * it adds collections.Mapping to the list of parent classes (i.e. to the class' `__bases__`)
    * it generates `__len__`, `__iter__` and `__getitem__` in order for the appropriate fields to be exposed in the dict
    view.
    * it adds a static from_dict method to build objects from dicts (only if only_constructor_args=True)
    * it overrides eq method if not already implemented
    * it overrides str and repr method if not already implemented
    
    Parameters allow to customize the list of fields that will be visible.

    :param include: a tuple of explicit attribute names to include (None means all)
    :param exclude: a tuple of explicit attribute names to exclude. In such case, include should be None.
    :param only_constructor_args: if True (default), only constructor arguments will be exposed through the dictionary
        view, not any other field that would be created in the constructor or dynamically. This makes it very convenient
        to use in combination with @autoargs. If set to False, the dictionary is a direct view of public object fields.
    :param only_public_fields: this parameter is only used when only_constructor_args is set to False. If
        only_public_fields is set to False, all fields are visible. Otherwise (default), class-private fields will be
        hidden
    :return:
    """
    return autodict_decorate(cls, include=include, exclude=exclude, only_constructor_args=only_constructor_args,
                             only_public_fields=only_public_fields)


def autodict_decorate(cls,                         # type: Type[T]
                      include=None,                # type: Union[str, Tuple[str]]
                      exclude=None,                # type: Union[str, Tuple[str]]
                      only_constructor_args=True,  # type: bool
                      only_public_fields=True      # type: bool
                      ):
    # type: (...) -> Type[T]
    """
    To automatically generate the appropriate methods so that objects of this class behave like a `dict`,
    manually, without using @autodict decorator.

    :param cls: the class on which to execute. Note that it won't be wrapped.
    :param include: a tuple of explicit attribute names to include (None means all)
    :param exclude: a tuple of explicit attribute names to exclude. In such case, include should be None.
    :param only_constructor_args: if True (default), only constructor arguments will be exposed through the dictionary
    view, not any other field that would be created in the constructor or dynamically. This makes it very convenient
    to use in combination with @autoargs. If set to False, the dictionary is a direct view of public object fields.
    :param only_public_fields: this parameter is only used when only_constructor_args is set to False. If
    only_public_fields is set to False, all fields are visible. Otherwise (default), class-private fields will be hidden
    :return:
    """

    # first check that we do not conflict with other known decorators
    _check_known_decorators(cls, '@autodict')

    # perform the class mod
    _execute_autodict_on_class(cls, include=include, exclude=exclude, only_constructor_args=only_constructor_args,
                               only_public_fields=only_public_fields)

    return cls


def _execute_autodict_on_class(object_type,                 # type: Type[T]
                               include=None,                # type: Union[str, Tuple[str]]
                               exclude=None,                # type: Union[str, Tuple[str]]
                               only_constructor_args=True,  # type: bool
                               only_public_fields=True      # type: bool
                               ):
    """
    This method makes objects of the class behave like a read-only `dict`. It does several things:
    * it adds collections.Mapping to the list of parent classes (i.e. to the class' `__bases__`)
    * it generates `__len__`, `__iter__` and `__getitem__` in order for the appropriate fields to be exposed in the dict
    view.
    * it adds a static from_dict method to build objects from dicts (only if only_constructor_args=True)
    * it overrides eq method if not already implemented
    * it overrides str and repr method if not already implemented

    Parameters allow to customize the list of fields that will be visible.

    :param object_type: the class on which to execute.
    :param include: a tuple of explicit attribute names to include (None means all)
    :param exclude: a tuple of explicit attribute names to exclude. In such case, include should be None.
    :param only_constructor_args: if True (default), only constructor arguments will be exposed through the dictionary
        view, not any other field that would be created in the constructor or dynamically. This makes it very convenient
        to use in combination with @autoargs. If set to False, the dictionary is a direct view of public object fields.
    :param only_public_fields: this parameter is only used when only_constructor_args is set to False. If
        only_public_fields is set to False, all fields are visible. Otherwise (default), class-private fields will be
        hidden
    :return:
    """
    # 0. first check parameters
    validate_include_exclude(include, exclude)

    # if issubclass(object_type, Mapping):
    #     raise ValueError('@autodict can not be set on classes that are already subclasses of Mapping, and therefore '
    #                      'already behave like dict')
    super_is_already_a_mapping = issubclass(object_type, Mapping)

    # 1. implement the abstract method required by Mapping to work, according to the options
    if only_constructor_args:
        # ** easy: we know the exact list of fields to make visible in the Mapping
        # a. Find the __init__ constructor signature
        constructor = get_constructor(object_type, allow_inheritance=True)
        s = signature(constructor)

        # b. Collect all attributes that are not 'self' and are included and not excluded
        added = []
        for attr_name in s.parameters.keys():
            if is_attr_selected(attr_name, include=include, exclude=exclude):
                added.append(attr_name)

        # c. Finally build the methods
        def __iter__(self):
            """
            Generated by @autodict.
            Implements the __iter__ method from collections.Iterable by relying on a hardcoded list of fields
            PLUS the super dictionary if relevant
            :param self:
            :return:
            """
            if super_is_already_a_mapping:
                return iter(added + [o for o in super(object_type, self).__iter__() if o not in added])
            else:
                return iter(added)

        # def __len__(self):
        #     """
        #     Generated by @autodict.
        #     Implements the __len__ method from collections.Sized by relying on a hardcoded list of fields
        #     PLUS the super dictionary if relevant
        #     :param self:
        #     :return:
        #     """
        #     if super_is_already_a_mapping:
        #         return len(added) + super(object_type, self).__len__()
        #     else:
        #         return len(added)

        if super_is_already_a_mapping:
            def __getitem__(self, key):
                """
                Generated by @autodict.
                Implements the __getitem__ method from collections.Mapping by relying on a hardcoded list of fields
                PLUS the parent dictionary when not found in self
                :param self:
                :param key:
                :return:
                """
                if key in added:
                    try:
                        return getattr(self, key)
                    except AttributeError:
                        try:
                            return super(object_type, self).__getitem__(key)
                        except Exception as e:
                            raise KeyError('@autodict generated dict view - {key} is a constructor parameter but is not'
                                           ' a field (was the constructor called ?). Delegating to super[{key}] raises '
                                           'an exception: {etyp} {err}'.format(key=key, etyp=type(e).__name__, err=e))
                else:
                    try:
                        return super(object_type, self).__getitem__(key)
                    except Exception as e:
                        raise KeyError('@autodict generated dict view - {key} is not a constructor parameter so not '
                                       ' handled by this dict view. Delegating to super[{key}] raised an exception: '
                                       '{etyp} {err}'.format(key=key, etyp=type(e).__name__, err=e))
        else:
            def __getitem__(self, key):
                """
                Generated by @autodict.
                Implements the __getitem__ method from collections.Mapping by relying on a hardcoded list of fields
                :param self:
                :param key:
                :return:
                """
                if key in added:
                    try:
                        return getattr(self, key)
                    except AttributeError:
                        raise KeyError('@autodict generated dict view - {} is a constructor parameter but is not a '
                                       'field (was the constructor called ?)'.format(key))
                else:
                    raise KeyError('@autodict generated dict view - invalid or hidden field name: %s' % key)

    else:
        # ** all dynamic fields are allowed
        if include is None and exclude is None and not only_public_fields:
            # easy: all of vars is exposed
            def __iter__(self):
                """
                Generated by @autodict.
                Implements the __iter__ method from collections.Iterable by relying on vars(self)
                PLUS the super dictionary if relevant
                :param self:
                :return:
                """
                if super_is_already_a_mapping:
                    return iter(list(vars(self)) + [o for o in super(object_type, self).__iter__()
                                                    if o not in vars(self)])
                else:
                    return iter(vars(self))

            # def __len__(self):
            #     """
            #     Generated by @autodict.
            #     Implements the __len__ method from collections.Sized by relying on vars(self)
            #     PLUS the super dictionary if relevant
            #     :param self:
            #     :return:
            #     """
            #     if super_is_already_a_mapping:
            #         return len(list(vars(self)) + [o for o in super(object_type, self).__iter__()
            #                                        if o not in vars(self)])
            #     else:
            #         return len(vars(self))

            if super_is_already_a_mapping:
                def __getitem__(self, key):
                    """
                    Generated by @autodict.
                    Implements the __getitem__ method from collections.Mapping by relying on getattr(self, key)
                    PLUS the super dictionary
                    :param self:
                    :param key:
                    :return:
                    """
                    try:
                        return getattr(self, key)
                    except AttributeError:
                        try:
                            return super(object_type, self).__getitem__(key)
                        except Exception as e:
                            raise KeyError('@autodict generated dict view - {key} is not a valid field (was the '
                                           'constructor called?). Delegating to super[{key}] raises an exception: '
                                           '{etyp} {err}'.format(key=key, etyp=type(e).__name__, err=e))
            else:
                def __getitem__(self, key):
                    """
                    Generated by @autodict.
                    Implements the __getitem__ method from collections.Mapping by relying on getattr(self, key)
                    :param self:
                    :param key:
                    :return:
                    """
                    try:
                        return getattr(self, key)
                    except AttributeError:
                        raise KeyError('@autodict generated dict view - {key} is not a valid field (was the '
                                       'constructor called?)'.format(key=key))
        else:
            # harder: all fields are allowed, but there are filters on this dynamic list
            # private_name_prefix = '_' + object_type.__name__ + '_'
            private_name_prefix = '_'

            if super_is_already_a_mapping:
                def __iter__(self):
                    """
                    Generated by @autodict.
                    Implements the __iter__ method from collections.Iterable by relying on a filtered vars(self)
                    :param self:
                    :return:
                    """
                    myattrs = [possibly_replace_with_property_name(self.__class__, att_name) for att_name in vars(self)]
                    for att_name in myattrs + [o for o in super(object_type, self).__iter__() if o not in vars(self)]:
                        if is_attr_selected(att_name, include=include, exclude=exclude):
                            if not only_public_fields \
                                    or (only_public_fields and not att_name.startswith(private_name_prefix)):
                                yield att_name
            else:
                def __iter__(self):
                    """
                    Generated by @autodict.
                    Implements the __iter__ method from collections.Iterable by relying on a filtered vars(self)
                    :param self:
                    :return:
                    """
                    for att_name in [possibly_replace_with_property_name(self.__class__, att_name)
                                     for att_name in vars(self)]:
                        if is_attr_selected(att_name, include=include, exclude=exclude):
                            if not only_public_fields \
                                    or (only_public_fields and not att_name.startswith(private_name_prefix)):
                                yield att_name

            # def __len__(self):
            #     """
            #     Generated by @autodict.
            #     Implements the __len__ method from collections.Sized by relying on a filtered vars(self)
            #     :param self:
            #     :return:
            #     """
            #     # rely on iter()
            #     return sum(1 for e in self)

            if super_is_already_a_mapping:
                def __getitem__(self, key):
                    """
                    Generated by @autodict.
                    Implements the __getitem__ method from collections.Mapping by relying on a filtered getattr(self, key)
                    :param self:
                    :param key:
                    :return:
                    """
                    if hasattr(self, key):
                        key = possibly_replace_with_property_name(self.__class__, key)
                        if is_attr_selected(key, include=include, exclude=exclude) and \
                                (not only_public_fields or
                                 (only_public_fields and not key.startswith(private_name_prefix))):
                            return getattr(self, key)
                        else:
                            try:
                                return super(object_type, self).__getitem__(key)
                            except Exception as e:
                                raise KeyError('@autodict generated dict view - {key} is a '
                                               'hidden field and super[{key}] raises an exception: {etyp} {err}'
                                               ''.format(key=key, etyp=type(e).__name__, err=e))
                    else:
                        try:
                            return super(object_type, self).__getitem__(key)
                        except Exception as e:
                            raise KeyError('@autodict generated dict view - {key} is an '
                                           'invalid field name (was the constructor called?). Delegating to '
                                           'super[{key}] raises an exception: {etyp} {err}'
                                           ''.format(key=key, etyp=type(e).__name__, err=e))
            else:
                def __getitem__(self, key):
                    """
                    Generated by @autodict.
                    Implements the __getitem__ method from collections.Mapping by relying on a filtered getattr(self, key)
                    :param self:
                    :param key:
                    :return:
                    """
                    if hasattr(self, key):
                        key = possibly_replace_with_property_name(self.__class__, key)
                        if is_attr_selected(key, include=include, exclude=exclude) and \
                                (not only_public_fields or
                                 (only_public_fields and not key.startswith(private_name_prefix))):
                            return getattr(self, key)
                        else:
                            raise KeyError('@autodict generated dict view - hidden field name: ' + key)
                    else:
                        raise KeyError('@autodict generated dict view - {key} is an invalid field name (was the '
                                       'constructor called? are the constructor arg names identical to the field '
                                       'names ?)'.format(key=key))

    def __len__(self):
        """
        Generated by @autodict.
        Implements the __len__ method from collections.Sized by relying on self.__iter__, so that the length will always
        match the true length.

        :param self:
        :return:
        """
        return sum(1 for e in self)

    if method_already_there(object_type, '__len__', this_class_only=True):
        if not hasattr(object_type.__len__, __AUTODICT_OVERRIDE_ANNOTATION):
            warn('__len__ is already defined on class {}, it will be overridden with the one generated by '
                 '@autodict/@autoclass ! If you want to use your version, annotate it with @autodict_override'
                 ''.format(str(object_type)))
            object_type.__len__ = __len__
    else:
        object_type.__len__ = __len__

    if method_already_there(object_type, '__iter__', this_class_only=True):
        if not hasattr(object_type.__iter__, __AUTODICT_OVERRIDE_ANNOTATION):
            warn('__iter__ is already defined on class {}, it will be overridden with the one generated by '
                 '@autodict/@autoclass ! If you want to use your version, annotate it with @autodict_override'
                 ''.format(str(object_type)))
            object_type.__iter__ = __iter__
    else:
        object_type.__iter__ = __iter__

    if method_already_there(object_type, '__getitem__', this_class_only=True):
        if not hasattr(object_type.__getitem__, __AUTODICT_OVERRIDE_ANNOTATION):
            warn('__getitem__ is already defined on class {}, it will be overridden with the one generated by '
                 '@autodict/@autoclass ! If you want to use your version, annotate it with @autodict_override'
                 ''.format(str(object_type)))
        else:
            object_type.__getitem__ = __getitem__
    else:
        object_type.__getitem__ = __getitem__

    # 2. add the methods from Mapping to the class
    # -- current proposition: add inheritance dynamically
    type_bases = object_type.__bases__
    if Mapping not in type_bases:
        bazz = tuple(t for t in type_bases if t is not object)
        if len(bazz) == len(type_bases):
            # object was not there
            new_bases = bazz + (Mapping,)
        else:
            # object was there, put it at the end
            new_bases = bazz + (Mapping, object)

        try:
            object_type.__bases__ = new_bases
        except TypeError:
            try:
                # maybe a metaclass issue, we can try this
                object_type.__bases__ = with_metaclass(type(object_type), *new_bases)
            except TypeError:
                # python 2.x and object type is a new-style class directly inheriting from object
                # open bug: https://bugs.python.org/issue672115

                # -- alternate way: add methods one by one
                names = [
                    # no need
                    # '__class__', '__metaclass__', '__subclasshook__', '__init__', '__ne__', '__new__'
                    # no need: object
                    # '__getattribute__','__delattr__','__setattr__','__format__','__reduce__','__reduce_ex__','__sizeof__'
                    # -----
                    # '__getitem__', overridden above
                    # '__iter__', overridden above
                    # '__len__', overridden above
                    # '__eq__', overridden below
                    # '__repr__',  overridden below
                    # '__str__', overridden below
                    '__contains__',
                    'get',
                    'items',
                    'iteritems',
                    'iterkeys',
                    'itervalues',
                    'keys',
                    'values']
                # from inspect import getmembers
                # def is_useful(m):
                #     return m
                # meths = getmembers(Mapping.get(), predicate=is_useful)
                # for name, func in meths:
                for name in names:
                    # bind method to this class too (we access 'im_func' to get the original method)
                    setattr(object_type, name, getattr(Mapping, name).im_func)

    # 3. add the static class method to build objects from a dict
    # if only_constructor_args:

    # only do it if there is no existing method on the type
    if not method_already_there(object_type, 'from_dict'):
        def from_dict(cls,
                      dct  # type: Dict[str, Any]
                      ):
            """
            Generated by @autodict.
            A class method to construct an object from a dictionary of field values.

            :param cls:
            :param dct:
            :return:
            """
            return cls(**dct)

        object_type.from_dict = classmethod(from_dict)

    # 4. override equality method if not already implemented LOCALLY (on this type - we dont care about the super
    # since we'll delegate to them when we can't handle)
    if not method_already_there(object_type, '__eq__', this_class_only=True):

        def __eq__(self, other):
            """
            Generated by @autodict.
            In the case the other is of the same type, use the dict comparison. Otherwise, falls back to super.

            :param self:
            :param other:
            :return:
            """
            # in the case the other is of the same type, use the dict comparison, that relies on the appropriate fields
            if isinstance(other, object_type):
                return dict(self) == dict(other)
            else:
                # else fallback to inherited behaviour, whatever it is
                try:
                    f = super(object_type, self).__eq__
                except AttributeError:
                    # can happen in python 2 when adding Mapping inheritance failed
                    return Mapping.__eq__(dict(self), other)
                else:
                    return f(other)

        object_type.__eq__ = __eq__

    # 5. override str and repr method if not already implemented
    if not method_already_there(object_type, '__str__', this_class_only=True):

        def __str__(self):
            """
            Generated by @autodict.
            Uses the dict representation and puts the type in front

            :param self:
            :return:
            """
            # python 2 compatibility: use self.__class__ not type()
            return self.__class__.__name__ + '(' + print_ordered_dict(self) + ')'

        object_type.__str__ = __str__

    if not method_already_there(object_type, '__repr__', this_class_only=True):
        def __repr__(self):
            """
            Generated by @autodict.
            Uses the dict representation and puts the type in front
            maybe?

            :param self:
            :return:
            """
            # python 2 compatibility: use self.__class__ not type()
            return self.__class__.__name__ + '(' + print_ordered_dict(self) + ')'

        object_type.__repr__ = __repr__

    return


def print_ordered_dict(obj):
    # This destroys the order
    # return str(dict(obj))

    # This follows the order from __iter__
    return '{' + ', '.join('{}: {}'.format(repr(k), repr(v)) for k, v in obj.items()) + '}'


def autodict_override_decorate(func  # type: Callable
                               ):
    # type: (...) -> Callable
    """
    Used to decorate a function as an overridden dictionary method (such as __iter__), without using the
    @autodict_override annotation.

    :param func: the function on which to execute. Note that it won't be wrapped but simply annotated.
    :return:
    """

    if func.__name__ not in {Mapping.__iter__.__name__, Mapping.__getitem__.__name__, Mapping.__len__.__name__}:
        raise ValueError('@autodict_override can only be used on one of the three Mapping methods __iter__,'
                         '__getitem__ and __len__. Found: ' + func.__name__)

    # Simply annotate the function
    if hasattr(func, __AUTODICT_OVERRIDE_ANNOTATION):
        raise DuplicateOverrideError('Function is overridden twice : ' + func.__name__)
    else:
        setattr(func, __AUTODICT_OVERRIDE_ANNOTATION, True)

    return func


autodict_override = autodict_override_decorate
"""A decorator to indicate an overridden dictionary method. In this case autodict will not override it and will not 
generate a warning"""



try:  # python 3.5+
    from typing import Union, Tuple, TypeVar
    try:  # python 3.5.3+
        from typing import Type
    except ImportError:
        pass
    T = TypeVar('T')
except ImportError:
    pass

from autoclass.autoargs_ import autoargs_decorate
from autoclass.autoprops_ import autoprops_decorate
from autoclass.autodict_ import autodict_decorate
from autoclass.utils import get_constructor
from autoclass.autohash_ import autohash_decorate


from decopatch import class_decorator, DECORATED


@class_decorator
def autoclass(include=None,    # type: Union[str, Tuple[str]]
              exclude=None,    # type: Union[str, Tuple[str]]
              autoargs=True,   # type: bool
              autoprops=True,  # type: bool
              autodict=True,   # type: bool
              autohash=True,   # type: bool
              cls=DECORATED
              ):
    """
    A decorator to perform @autoargs, @autoprops and @autodict all at once with the same include/exclude list.

    :param include: a tuple of explicit attribute names to include (None means all)
    :param exclude: a tuple of explicit attribute names to exclude. In such case, include should be None.
    :param autoargs: a boolean to enable autoargs on the consturctor (default: True)
    :param autoprops: a boolean to enable autoargs on the consturctor (default: True)
    :param autodict: a boolean to enable autoargs on the consturctor (default: True)
    :param autohash: a boolean to enable autohash on the constructor (default: True)
    :return:
    """
    return autoclass_decorate(cls, include=include, exclude=exclude, autoargs=autoargs, autoprops=autoprops,
                              autodict=autodict, autohash=autohash)


def autoclass_decorate(cls,             # type: Type[T]
                       include=None,    # type: Union[str, Tuple[str]]
                       exclude=None,    # type: Union[str, Tuple[str]]
                       autoargs=True,   # type: bool
                       autoprops=True,  # type: bool
                       autodict=True,   # type: bool
                       autohash=True    # type: bool
                       ):
    # type: (...) -> Type[T]
    """

    :param cls: the class on which to execute. Note that it won't be wrapped.
    :param include: a tuple of explicit attributes to include (None means all)
    :param exclude: a tuple of explicit attribute names to exclude. In such case, include should be None.
    :param autoargs: a boolean to enable autoargs on the consturctor (default: True)
    :param autoprops: a boolean to enable autoprops on the consturctor (default: True)
    :param autodict: a boolean to enable autodict on the consturctor (default: True)
    :param autohash: a boolean to enable autohash on the constructor (default: True)
    :return:
    """

    # @autoargs
    if autoargs:
        init = get_constructor(cls)
        cls.__init__ = autoargs_decorate(init, include=include, exclude=exclude)

    # @autoprops
    if autoprops:
        autoprops_decorate(cls, include=include, exclude=exclude)

    # @autodict
    if autodict:
        autodict_decorate(cls, include=include, exclude=exclude)

    # @autohash
    if autohash:
        autohash_decorate(cls, include=include, exclude=exclude)

    return cls



from autoclass.autoargs_ import autoargs, autoargs_decorate
from autoclass.autoclass_ import autoclass, autoclass_decorate
from autoclass.autodict_ import autodict, autodict_decorate, autodict_override, autodict_override_decorate, \
    print_ordered_dict
from autoclass.autohash_ import autohash, autohash_decorate
from autoclass.autoprops_ import IllegalGetterSignatureException, IllegalSetterSignatureException, autoprops, \
    autoprops_decorate, DuplicateOverrideError, getter_override, setter_override, autoprops_override_decorate
from autoclass.utils import AutoclassDecorationException

__all__ = [
    # submodules
    'autoargs_', 'autoclass_', 'autodict_', 'autohash_', 'autoprops_', 'utils',
    # symbols
    'autoargs', 'autoargs_decorate',
    'autoclass', 'autoclass_decorate',
    'autodict', 'autodict_decorate', 'autodict_override', 'autodict_override_decorate', 'print_ordered_dict',
    'autohash', 'autohash_decorate',
    'IllegalGetterSignatureException', 'IllegalSetterSignatureException', 'autoprops', 'autoprops_decorate',
    'DuplicateOverrideError', 'getter_override', 'setter_override', 'autoprops_override_decorate',
    'AutoclassDecorationException'
]



try:
    from inspect import signature
except ImportError:
    from funcsigs import signature

try:
    from typing import Tuple, Union, TypeVar
    try:
        from typing import Type
    except ImportError:
        pass
    T = TypeVar('T')

except ImportError:
    pass

from decopatch import class_decorator, DECORATED

from autoclass.utils import is_attr_selected, method_already_there, possibly_replace_with_property_name, \
    validate_include_exclude
from autoclass.utils import get_constructor
from autoclass.utils import _check_known_decorators


@class_decorator
def autohash(include=None,                # type: Union[str, Tuple[str]]
             exclude=None,                # type: Union[str, Tuple[str]]
             only_constructor_args=False,  # type: bool
             only_public_fields=False,     # type: bool
             cls=DECORATED
             ):
    """
    A decorator to makes objects of the class implement __hash__, so that they can be used correctly for example in
    sets.
    
    Parameters allow to customize the list of attributes that are taken into account in the hash.

    :param include: a tuple of explicit attribute names to include (None means all)
    :param exclude: a tuple of explicit attribute names to exclude. In such case, include should be None.
    :param only_constructor_args: if False (default), all fields will be included in the hash, even if they are defined
        in the constructor or dynamically. If True, only constructor arguments will be included in the hash, not any
        other field that would be created in the constructor or dynamically. Please note that this behaviour is the
        opposite from @autodict.
    :param only_public_fields: this parameter is only used when only_constructor_args is set to False. If
        only_public_fields is set to False (default), all fields are used in the hash. Otherwise, class-private fields
        will not be taken into account in the hash. Please note that this behaviour is the opposite from @autodict.
    :return:
    """

    return autohash_decorate(cls, include=include, exclude=exclude, only_constructor_args=only_constructor_args,
                             only_public_fields=only_public_fields)


def autohash_decorate(cls,                          # type: Type[T]
                      include=None,                 # type: Union[str, Tuple[str]]
                      exclude=None,                 # type: Union[str, Tuple[str]]
                      only_constructor_args=False,  # type: bool
                      only_public_fields=False,     # type: bool
                      ):
    # type: (...) -> Type[T]
    """
    To automatically generate the appropriate methods so that objects of this class are hashable,
    manually, without using @autohash decorator.

    :param cls: the class on which to execute. Note that it won't be wrapped.
    :param include: a tuple of explicit attribute names to include (None means all)
    :param exclude: a tuple of explicit attribute names to exclude. In such case, include should be None.
    :param only_constructor_args: if False (default), all fields will be included in the hash, even if they are defined
    in the constructor or dynamically. If True, only constructor arguments will be included in the hash, not any other
    field that would be created in the constructor or dynamically. Please note that this behaviour is the opposite from
    @autodict.
    :param only_public_fields: this parameter is only used when only_constructor_args is set to False. If
    only_public_fields is set to False (default), all fields are used in the hash. Otherwise, class-private fields will
    not be taken into account in the hash. Please note that this behaviour is the opposite from @autodict.
    :return:
    """

    # first check that we do not conflict with other known decorators
    _check_known_decorators(cls, '@autohash')

    # perform the class mod
    _execute_autohash_on_class(cls, include=include, exclude=exclude, only_constructor_args=only_constructor_args,
                               only_public_fields=only_public_fields)

    return cls


def _execute_autohash_on_class(object_type,                  # type: Type[T]
                               include=None,                 # type: Union[str, Tuple[str]]
                               exclude=None,                 # type: Union[str, Tuple[str]]
                               only_constructor_args=False,  # type: bool
                               only_public_fields=False,     # type: bool
                               ):
    """
    A decorator to make objects of the class implement __hash__, so that they can be used correctly for example in
    sets.

    Parameters allow to customize the list of attributes that are taken into account in the hash.

    :param object_type: the class on which to execute.
    :param include: a tuple of explicit attribute names to include (None means all)
    :param exclude: a tuple of explicit attribute names to exclude. In such case, include should be None.
    :param only_constructor_args: if False (default), all fields will be included in the hash, even if they are defined
    in the constructor or dynamically. If True, only constructor arguments will be included in the hash, not any other
    field that would be created in the constructor or dynamically. Please note that this behaviour is the opposite from
    @autodict.
    :param only_public_fields: this parameter is only used when only_constructor_args is set to False. If
    only_public_fields is set to False (default), all fields are used in the hash. Otherwise, class-private fields will
    not be taken into account in the hash. Please note that this behaviour is the opposite from @autodict.
    :return:
    """
    # First check parameters
    validate_include_exclude(include, exclude)

    # Override hash method if not already implemented
    if not method_already_there(object_type, '__hash__'):
        if only_constructor_args:
            # a. Find the __init__ constructor signature
            constructor = get_constructor(object_type, allow_inheritance=True)
            s = signature(constructor)

            # b. Collect all attributes that are not 'self' and are included and not excluded
            added = []
            # we assume that the order of attributes will always be the same here....
            for attr_name in s.parameters.keys():
                if is_attr_selected(attr_name, include=include, exclude=exclude):
                    added.append(attr_name)

            # c. Finally build the method
            def __hash__(self):
                """
                Generated by @autohash.
                Implements the __hash__ method by hashing a tuple of selected attributes
                :param self:
                :return:
                """
                # note: we prepend a unique hash for the class  > NO, it is more intuitive to not do that.
                # return hash(tuple([type(self)] + [getattr(self, att_name) for att_name in added]))
                return hash(tuple(getattr(self, att_name) for att_name in added))

        else:
            # ** all dynamic fields are allowed
            if include is None and exclude is None and not only_public_fields:

                # easy: all of vars values is included in the hash
                def __hash__(self):
                    """
                    Generated by @autohash.
                    Implements the __hash__ method by hashing vars(self).values()
                    :param self:
                    :return:
                    """
                    # note: we prepend a unique hash for the class  > NO, it is more intuitive to not do that.
                    # return hash(tuple([type(self)] + list(vars(self).values())))
                    return hash(tuple(vars(self).values()))

            else:
                # harder: dynamic filters
                # private_name_prefix = '_' + object_type.__name__ + '_'
                private_name_prefix = '_'

                def __hash__(self):
                    """
                    Generated by @autohash.
                    Implements the __hash__ method by hashing the tuple of included/not excluded field values, possibly
                    not including the private fields if `only_public_fields` was set to True

                    :param self:
                    :return:
                    """
                    # note: we prepend a unique hash for the class > NO, it is more intuitive to not do that.
                    # to_hash = [type(self)]
                    to_hash = []

                    for att_name, att_value in vars(self).items():
                        att_name = possibly_replace_with_property_name(self.__class__, att_name)
                        if is_attr_selected(att_name, include=include, exclude=exclude):
                            if not only_public_fields \
                                    or (only_public_fields and not att_name.startswith(private_name_prefix)):
                                to_hash.append(att_value)

                    return hash(tuple(to_hash))

        # now set the method on the class
        object_type.__hash__ = __hash__

    return
